/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2005 Intel Corporation 
All rights reserved. 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */

//mlsTraceAssist.h
//Contains globals and utility procedures

// Bugs and criticisms can be directed to
// myers.albert@gmail.com

//KNOB types used to parse tool parameters
KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE, "pintool", 
                    "o",      "mlsTrace.out",    "specify trace output file name");
KNOB<BOOL> ASCII(KNOB_MODE_WRITEONCE, "pintool", 
                    "a",      "0",               "use ASCII output file instead of binary");
KNOB<UINT64> Skip(KNOB_MODE_WRITEONCE,"pintool", 
                    "s",      "0",               "Begin emitting descriptors after executing a specified number of instructions ");
KNOB<UINT64> Length(KNOB_MODE_WRITEONCE, "pintool", 
                    "l",      "0",               "Number of instructions to profile (default is no limit)");
KNOB<std::string> Compression(KNOB_MODE_WRITEONCE, "pintool",
                    "c",      "0",               "Compress trace. Supports bzip2, pbzip2, gzip, and pigz");
KNOB<BOOL> AnnotateDisassembly(KNOB_MODE_WRITEONCE, "pintool", 
                    "d",      "0",               "Annotate descriptors with disassembly (only works when output is ASCII");
KNOB<UINT64> FileLimit(KNOB_MODE_WRITEONCE, "pintool", 
                    "f",      "50000",           "Output file size limit in MB. Tracing will end after reaching this limit. Default is 50000 MB");

KNOB<BOOL> TraceLoad(KNOB_MODE_WRITEONCE, "pintool",
                    "load",   "1",               "Trace load instructions (default)");
KNOB<BOOL> TraceStore(KNOB_MODE_WRITEONCE, "pintool",
                    "store", "0",                "Trace store instructions (default)");

//holds the number of threads
UINT8 numThreads;
static TLS_KEY tls_key;
//get local thread id
UINT8 get_localtid(THREADID threadid)
{
   UINT8* localtid = static_cast<UINT8*>(PIN_GetThreadData(tls_key, threadid));
   return *localtid;
}
//protects descriptor table
PIN_LOCK table_lock;
//protects xed decoder
PIN_LOCK decoder_lock;
//protects streams
PIN_LOCK cout_lock;
//protects statistics globals
PIN_LOCK count_lock;
//protects memory operations by pin
//also used to protect store instructions
//store values are received by inspecting the 
//memory location after the instruction executes,
//but the process needs to be atomic, since another thread
//could change the value before it is inspected
PIN_LOCK mem_lock;

//Globals for file
struct tm stamp;
std::string fileName;
std::string statFileName;
//Used to find pin tool overhead time.
timeval t1, t2;

//ofstream for raw files
std::ofstream OutFile;
//FILE for piping to compression utility
FILE *outPipe;

//output file size limit;
UINT64 fileCount;
UINT64 fileLimit;
//number of instructions to skip;
UINT64  fastforwardLength; 
//number of instructions to profile
UINT64 traceLength;
//if a length isn't specififed
BOOL noLength;
//if using compression
BOOL usingCompression;

//Total number of load operands
//automatically initialized to zero
//holds operand count
//first index is loads, second is stores
UINT64 numTotalOperands[2];
UINT64 numByteOperands[2]; //1 byte
UINT64 numWordOperands[2]; //2 bytes
UINT64 numDoublewordOperands[2]; //4 bytes
UINT64 numQuadwordOperands[2]; //8 bytes
UINT64 numOctawordOperands[2]; //16 bytes
UINT64 numHexawordOperands[2]; //32 bytes
UINT64 numExtendedPrecisionOperands[2]; //10 bytes
UINT64 numOtherOperands[2]; //any other operand sizes

typedef enum : UINT8
{
   load  = 0,
   store = 1,
} traceType;
//Bin struct
typedef struct {
   UINT8 tid;
   traceType type;
   ADDRINT insAddr;
   intptr_t operandEffAddr;
   UINT8 operandSize;
   UINT8 *data;
} BinaryDescriptorTableEntry;
UINT16 BinaryDescriptorSize = 3*sizeof(UINT8)+sizeof(traceType)+2*sizeof(ADDRINT); 
                             
//holds load instruction descriptors
std::deque<BinaryDescriptorTableEntry> binDescriptorTable;
std::deque<std::string> ASCIIDescriptorTable;

//write ASCII descriptor table to file
inline VOID WriteASCIIDescriptorTableToFile(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   //write contents to disk
   while( !ASCIIDescriptorTable.empty() )
   {
      OutFile << ASCIIDescriptorTable.front();
      ASCIIDescriptorTable.pop_front();
   }
   ReleaseLock(&table_lock);
}

//write ASCII descriptor table to compression program
inline VOID WriteASCIIDescriptorTableToCompressor(const THREADID threadid)
{
   GetLock(&table_lock, threadid+1);
   //write contents to disk
   while( !ASCIIDescriptorTable.empty() )
   {
      fputs(ASCIIDescriptorTable.front().c_str(), outPipe);
      ASCIIDescriptorTable.pop_front();
   }
   ReleaseLock(&table_lock);
}

//write binary descriptor table to file
inline VOID WriteBinaryDescriptorTableToFile(const THREADID threadid)
{
      GetLock(&table_lock, threadid+1);
      while( !binDescriptorTable.empty() ) 
      {
         BinaryDescriptorTableEntry temp = binDescriptorTable.front();
         OutFile.write((char *)&temp.tid, sizeof(temp.tid));
         if( TraceLoad.Value() && TraceStore.Value() )
            OutFile.write((char *)&temp.type, sizeof(temp.type));
         OutFile.write((char *)&temp.insAddr, sizeof(temp.insAddr));
         OutFile.write((char *)&temp.operandEffAddr, sizeof(temp.operandEffAddr));
         OutFile.write((char *)&temp.operandSize, sizeof(temp.operandSize));
         OutFile.write((char *)temp.data, temp.operandSize);
         //delete allocated memory
         delete[] temp.data;
         binDescriptorTable.pop_front();
      }
      ReleaseLock(&table_lock);
}
//write binary descriptor table to compression program
inline VOID WriteBinaryDescriptorTableToCompressor(const THREADID threadid)
{
      GetLock(&table_lock, threadid+1);
      while( !binDescriptorTable.empty() ) 
      {
         BinaryDescriptorTableEntry temp = binDescriptorTable.front();
         fwrite(&temp.tid, sizeof(temp.tid), 1,outPipe);
         if( TraceLoad.Value() && TraceStore.Value() )
            fwrite(&temp.type, sizeof(temp.type), 1, outPipe);
         fwrite(&temp.insAddr, sizeof(temp.insAddr), 1, outPipe);
         fwrite(&temp.operandEffAddr, sizeof(temp.operandEffAddr), 1,outPipe);
         fwrite(&temp.operandSize, sizeof(temp.operandSize), 1,outPipe);
         fwrite(temp.data, temp.operandSize, 1,outPipe);
         //delete allocated memory
         delete[] temp.data;
         binDescriptorTable.pop_front();
      }
      ReleaseLock(&table_lock);
}
//Worker ASCII write thread when not using compression
VOID ThreadWriteASCII(VOID *arg)
{
   THREADID threadid = PIN_ThreadId();
   if(usingCompression)
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteASCIIDescriptorTableToCompressor(threadid);
      }
   }
   else
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteASCIIDescriptorTableToFile(threadid);
      }
   }
}

//Worker bin write thread when not using compression
VOID ThreadWriteBin(VOID *arg)
{
   THREADID threadid = PIN_ThreadId();
   if(usingCompression)
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteBinaryDescriptorTableToCompressor(threadid);
      }
   }
   else
   {
      while(1)
      {
         //if process is closing (entered fini()) kill thread
         if( PIN_IsProcessExiting() )
            PIN_ExitThread(1);

         WriteBinaryDescriptorTableToFile(threadid);
      }
   }
}

//Used to check fast forward and trace length in analysis functions
inline BOOL CanEmit(const THREADID threadid)
{
   GetLock(&count_lock, threadid+1);

   if( fastforwardLength < Skip.Value()  )
   {
      ReleaseLock(&count_lock);
      return false;
   }
   else
   {
      ReleaseLock(&count_lock);
      return true; 
   }
}
//check file size and exit early if too large
inline VOID IncrementFileCount(UINT32 size)
{
   //increment counter
   fileCount += size;
   //if limit reached, exit
   if(fileCount >= fileLimit)
      PIN_Detach();
}

//Address width
//extended
#if defined(TARGET_IA32E)
   #define ADDRWID 16
//32 bit
#else
   #define ADDRWID 8
#endif

//ins contains the instruction string
inline VOID getAssemblyString(string & ins, const THREADID threadid, const ADDRINT address)
{
   GetLock(&decoder_lock, threadid+1);

   //Use Intel XED to decode the instruction based on its address
#if defined(TARGET_IA32E)
    static const xed_state_t dstate = {XED_MACHINE_MODE_LONG_64, XED_ADDRESS_WIDTH_64b};
#else
    static const xed_state_t dstate = { XED_MACHINE_MODE_LEGACY_32, XED_ADDRESS_WIDTH_32b};
#endif
    xed_decoded_inst_t xedd;
    xed_decoded_inst_zero_set_mode(&xedd,&dstate);

   unsigned int max_inst_len = 15;

    xed_error_enum_t xed_code = xed_decode(&xedd, reinterpret_cast<UINT8*>(address), max_inst_len);
    BOOL xed_ok = (xed_code == XED_ERROR_NONE);
    if (xed_ok) {
        char buf[2048];
        xed_uint64_t runtime_address = static_cast<xed_uint64_t>(address);
        xed_decoded_inst_dump_intel_format(&xedd, buf, 2048, runtime_address);

        ins.assign(buf);
    }
    else
        ins = "XED failed to decode this instruction";

   ReleaseLock(&decoder_lock);
} 

inline void IncrementLoadStatistics(const UINT8 opSize)
{
   //atomics
   __sync_fetch_and_add(numTotalOperands+load, 1);
   switch(opSize)
   {
      case 1:
         __sync_fetch_and_add(numByteOperands+load, 1);
         break;
      case 2:
         __sync_fetch_and_add(numWordOperands+load, 1);
         break;
      case 4:
         __sync_fetch_and_add(numDoublewordOperands+load, 1);
         break;
      case 8:
         __sync_fetch_and_add(numQuadwordOperands+load, 1);
         break;
      case 10:
         __sync_fetch_and_add(numExtendedPrecisionOperands+load, 1);
         break;
      case 16:
         __sync_fetch_and_add(numOctawordOperands+load, 1);
         break;
      case 32:
         __sync_fetch_and_add(numHexawordOperands+load, 1);
         break;
      default:
         __sync_fetch_and_add(numOtherOperands+load, 1);
         break;
   }
}
inline void IncrementStoreStatistics(const UINT8 opSize)
{
   //atomics
   __sync_fetch_and_add(numTotalOperands+store, 1);
   switch(opSize)
   {
      case 1:
         __sync_fetch_and_add(numByteOperands+store, 1);
         break;
      case 2:
         __sync_fetch_and_add(numWordOperands+store, 1);
         break;
      case 4:
         __sync_fetch_and_add(numDoublewordOperands+store, 1);
         break;
      case 8:
         __sync_fetch_and_add(numQuadwordOperands+store, 1);
         break;
      case 10:
         __sync_fetch_and_add(numExtendedPrecisionOperands+store, 1);
         break;
      case 16:
         __sync_fetch_and_add(numOctawordOperands+store, 1);
         break;
      case 32:
         __sync_fetch_and_add(numHexawordOperands+store, 1);
         break;
      default:
         __sync_fetch_and_add(numOtherOperands+store, 1);
         break;
   }
}
//return percentage
inline double Percentage(const UINT64 num, const UINT64 den)
{
   if( den == 0 )
      return std::numeric_limits<double>::quiet_NaN();
   double ratio = (double)num/(double)den;
   return 100*ratio;
}

VOID PrintStatistics()
{
   std::ofstream StatFile;
   StatFile.open(fileName + ".Statistics");
   double elapsedTime;
   elapsedTime = (t2.tv_sec - t1.tv_sec) * 1000.0;
   elapsedTime += (t2.tv_usec - t1.tv_usec) / 1000.0;
  
   StatFile << "Instrumentation Time: " <<  elapsedTime << " ms" << std::endl;
   StatFile << "Number of Threads: " << static_cast<UINT32>(numThreads) << std::endl;
   StatFile << "Traced " << traceLength << " instructions" << std::endl;
   StatFile << "Skipped " << fastforwardLength << " instructions" << std::endl;  

  
   std::ostringstream bench;
   bench << static_cast<UINT32>(numThreads) << "," << traceLength << ",";

   if( TraceLoad.Value() )
   {
      StatFile << "Total Load Operands: " << numTotalOperands[0] << std::endl;
      StatFile.setf(std::ios::fixed);
      //1 byte
      StatFile << "\t" << numByteOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numByteOperands[0], numTotalOperands[0]) << " ) Byte Operands" << std::endl;
      //2 bytes
      StatFile << "\t" << numWordOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numWordOperands[0], numTotalOperands[0]) << " ) Word Operands" << std::endl;
      //4 bytes
      StatFile << "\t" << numDoublewordOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numDoublewordOperands[0], numTotalOperands[0]) << " ) Doubleword Operands" << std::endl;
      //8 bytes
      StatFile << "\t" << numQuadwordOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numQuadwordOperands[0], numTotalOperands[0]) << " ) Quadword Operands" << std::endl;
      //10 bytes
      StatFile << "\t" << numExtendedPrecisionOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numExtendedPrecisionOperands[0], numTotalOperands[0]) << " ) Extended Precision Operands" << std::endl;
      //16 bytes
      StatFile << "\t" << numOctawordOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numOctawordOperands[0], numTotalOperands[0]) << " ) Octaword Operands" << std::endl;
      //32 bytes
      StatFile << "\t" << numHexawordOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numHexawordOperands[0], numTotalOperands[0]) << " ) Hexaword Operands" << std::endl;
      //other sizes  
      StatFile << "\t" << numOtherOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numOtherOperands[0], numTotalOperands[0]) << " ) Operands of other size" << std::endl;

      //Albert 
      //used for benchmark characterizationa
      //bench << numTotalOperands[0] << "," << numByteOperands[0] << "," << numWordOperands[0] << "," << numDoublewordOperands[0] << "," << numQuadwordOperands[0]
      //                         << "," << numExtendedPrecisionOperands[0] << "," << numOctawordOperands[0] << "," << numHexawordOperands[0] << "," << numOtherOperands[0];
      
   }
   if( TraceStore.Value() )
   {
      StatFile << "Total Store Operands: " << numTotalOperands[1] << std::endl;
      StatFile.setf(std::ios::fixed);

      //1 byte
      StatFile << "\t" << numByteOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numByteOperands[1], numTotalOperands[1]) << " ) Byte Operands" << std::endl;
      //2 bytes
      StatFile << "\t" << numWordOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numWordOperands[1], numTotalOperands[1]) << " ) Word Operands" << std::endl;
      //4 bytes
      StatFile << "\t" << numDoublewordOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numDoublewordOperands[1], numTotalOperands[1]) << " ) Doubleword Operands" << std::endl;
      //8 bytes
      StatFile << "\t" << numQuadwordOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numQuadwordOperands[1], numTotalOperands[1]) << " ) Quadword Operands" << std::endl;
      //10 bytes
      StatFile << "\t" << numExtendedPrecisionOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numExtendedPrecisionOperands[1], numTotalOperands[1]) << " ) Extended Precision Operands" << std::endl;
      //16 bytes
      StatFile << "\t" << numOctawordOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numOctawordOperands[1], numTotalOperands[1]) << " ) Octaword Operands" << std::endl;
      //32 bytes
      StatFile << "\t" << numHexawordOperands[1] << " ( %" << std::setprecision(2)
                       << Percentage(numHexawordOperands[1], numTotalOperands[1]) << " ) Hexaword Operands" << std::endl;
      //other sizes  
      StatFile << "\t" << numOtherOperands[0] << " ( %" << std::setprecision(2)
                       << Percentage(numOtherOperands[1], numTotalOperands[1]) << " ) Operands of other size" << std::endl;
      //Albert 
      //used for benchmark characterizationa
      //bench << ", " << numTotalOperands[1] << "," << numByteOperands[1] << "," << numWordOperands[1] << "," << numDoublewordOperands[1] << "," << numQuadwordOperands[1]
      //                          << "," << numExtendedPrecisionOperands[1] << "," << numOctawordOperands[1] << "," << numHexawordOperands[1] << "," << numOtherOperands[1];

   }
   //Albert
   //StatFile << bench.str() << std::endl;
   //File Size
   //struct stat st;
   //stat(statFileName.c_str(), &st);
   //StatFile << st.st_size << std::endl;

   //StatFile << static_cast<UINT32>(numThreads) << "," << elapsedTime << std::endl;


}
VOID openOutput()
{

   //Check compression knob
   string compressor = Compression.Value();
   if(compressor != "0" && compressor != "bzip2" && compressor != "pbzip2" && compressor != "gzip" && compressor != "pigz")
   {
      std::cout << "mlsTrace: That compression program is not supported" << std::endl;
      std::exit(-1);
   }
   if(compressor != "0")
   {
      //const string temp = "/bin/" + compressor;
      if( access( string( "/bin/" + compressor ).c_str(), X_OK) == -1  && access( string( "/usr/bin/" + compressor ).c_str(), X_OK) == -1 ) 
      {
         std::cout << "mlsTrace: This system does not support " << compressor << ". Exiting..." << std::endl;
         std::exit(-1);
      }
      usingCompression = TRUE;
   }
   else
      usingCompression = FALSE;

   //attach extension
   string tempFileName;
   if(ASCII.Value() == FALSE)
      tempFileName = fileName + ".bin";
   else
      tempFileName = fileName + ".txt";
  

   if( usingCompression )
   {
      char temp[1000] = { 0 };
      if( compressor == "bzip2" )
      {
         std::cout<<"Test!"<<std::endl; 
         tempFileName += ".bz2"; 
         sprintf(temp, "bzip2 > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else if( compressor == "pbzip2" )
      {
         tempFileName += ".bz2"; 
         sprintf(temp, "pbzip2 > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else if( compressor == "gzip" )
      {
         tempFileName += ".gz"; 
         sprintf(temp, "gzip > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
      else
      {
         tempFileName += ".gz"; 
         sprintf(temp, "pigz > %s", tempFileName.c_str());
         outPipe = popen(temp, "w");
      }
   
      if( outPipe == NULL )
      { 
         std::cout << "mlsTrace: Error opening outPipe..." << std::endl;
         std::exit(-1);
      }
   }
   //if not using compression, write to raw file
   else
   {
      if( ASCII.Value() == FALSE )
         OutFile.open(tempFileName, ios::out | ios::binary);
      else
         OutFile.open(tempFileName);
   }
   statFileName = tempFileName;

   //if tracing load and store instructions need to write meta data for binary file
   //0 - only load or store
   //1 - both load and store
   if( !ASCII.Value() )
   {
      UINT8 meta;
      if( TraceLoad.Value() && TraceStore.Value() )
         meta = 1;
      else
         meta = 0;    

      if( usingCompression )
         fwrite(&meta, sizeof(UINT8), 1, outPipe);
      else
         OutFile.write((char *)&meta, sizeof(UINT8));
   }

   THREADID err;
   //Spawn write thread
   if( ASCII.Value() == FALSE )
      err = PIN_SpawnInternalThread(ThreadWriteBin, NULL, 0, NULL);
   else
      err = PIN_SpawnInternalThread(ThreadWriteASCII, NULL, 0, NULL);
 
   if( err==INVALID_THREADID )
      std::cout << "Error creating internal Pin thread. Exiting.." << std::endl;
}
VOID closeOutput()
{
   if(usingCompression) 
      pclose(outPipe);
   else
      OutFile.close();

   ASCIIDescriptorTable.clear();
   binDescriptorTable.clear();
}
